;; Generated package description from consult-hoogle.el  -*- no-byte-compile: t -*-
(define-package "consult-hoogle" "0.4.1" "Hoogle frontend using consult" '((emacs "28.1") (consult "2.0")) :commit "5b16f2f6c8c9c81a2539ba8371db1f76e6e4cc68" :authors '(("rahguzar" . "rahguzar@mailbox.org")) :maintainer '("rahguzar" . "rahguzar@mailbox.org") :keywords '("docs" "languages") :url "https://codeberg.org/rahguzar/consult-hoogle")
